#ifndef CPSE_MATRIX_H
#define CPSE_MATRIX_H

/** \file matrix.h
  * \brief Class definitions for matrix and vector.
  *
  * Include matrix.tpp for operations. */

#include <cpse/math/mathdefs.h>

namespace cpse {
namespace math {

class float_keeper;
template<class D, class S> struct assign_dispatch;

// matrix storage. parameterized by type so that it can be used by other
// types (e.g. cpse::basic_point)
template<int Rows, int Cols, class Type>
class basic_matrix_storage {
	Type m_data[Cols][Rows]; 
public:
	Type &operator()(int r, int c) { return m_data[c][r]; }
	Type operator()(int r, int c) const { return m_data[c][r]; }
	Type *data() { return &m_data[0][0]; }
	const Type *data() const { return &m_data[0][0]; }
};
// specializations for vectors
template<int Rows, class Type>
class basic_matrix_storage<Rows, 1, Type> {
	Type m_data[Rows];
public:
	Type &operator()(int r, int) { return m_data[r]; }
	Type operator()(int r, int) const { return m_data[r]; }
	Type &operator()(int r) { return m_data[r]; }
	Type operator()(int r) const { return m_data[r]; }
	Type *data() { return m_data; }
	const Type *data() const { return m_data; }
};
template<class Type>
class basic_matrix_storage<2, 1, Type> {
public:
	union { Type x; Type s; };
	union { Type y; Type t; };

	Type &operator()(int r, int) { return (&x)[r]; }
	Type operator()(int r, int) const { return (&x)[r]; }
	Type &operator()(int r) { return (&x)[r]; }
	Type operator()(int r) const { return (&x)[r]; }
	Type *data() { return &x; }
	const Type *data() const { return &x; }
};
template<class Type>
class basic_matrix_storage<3, 1, Type> {
public:
	union { Type x; Type s; };
	union { Type y; Type t; };
	union { Type z; Type u; };

	Type &operator()(int r, int) { return (&x)[r]; }
	Type operator()(int r, int) const { return (&x)[r]; }
	Type &operator()(int r) { return (&x)[r]; }
	Type operator()(int r) const { return (&x)[r]; }
	Type *data() { return &x; }
	const Type *data() const { return &x; }
};
template<class Type>
class basic_matrix_storage<4, 1, Type> {
public:
	union { Type x; Type s; };
	union { Type y; Type t; };
	union { Type z; Type u; };
	union { Type w; Type v; };

	Type &operator()(int r, int) { return (&x)[r]; }
	Type operator()(int r, int) const { return (&x)[r]; }
	Type &operator()(int r) { return (&x)[r]; }
	Type operator()(int r) const { return (&x)[r]; }
	Type *data() { return &x; }
	const Type *data() const { return &x; }
};

// the type most matrix operations work with. subclass tmatrix<basic_matrix>
// to make your own matrix type
template<int Rows, int Cols>
class basic_matrix : public basic_matrix_storage<Rows, Cols, float> {
public:
	static const int rows = Rows;
	static const int columns = Cols;
	static const bool is_matrix = true;
	static const bool is_scalar = false;
	static const bool error = false;
	static const bool element_op = true;
	typedef basic_matrix<Rows, Cols> base_type;

	basic_matrix() {}
	// used for implicit evaluation by the select_* templates in
	// matrix_operators.tpp
	template<class T> basic_matrix(const T &m);
	template<int I> float element() const { return data()[I]; }
};

// wrapped around matrix expressions by tmatrix_type to make overloading based
// on size possible. the type most operators accept
template<int Rows, int Cols, class Base>
class tmatrix_dim : public Base {
public:
	tmatrix_dim() {}
	template<class A> tmatrix_dim(const A &a) : Base(a) {}
	template<class A, class B> tmatrix_dim(const A &a, const B &b)
		: Base(a, b) {}
};

// should be wrapped around matrix expressions that are strictly for
// initialization to avoid having to write two constructor/assignment
// pairs per initializer
template<class Base>
class tmatrix_ndim : public Base {
public:
	tmatrix_ndim() {}
	template<class A> tmatrix_ndim(const A &a) : Base(a) {}
	template<class A, class B> tmatrix_ndim(const A &a, const B &b)
		: Base(a, b) {}
};

// wrapped around expressions by tmatrix. used for early detection of
// illegal operations
template<bool Is_matrix, bool Is_scalar, bool Element_op, class Base>
class tmatrix_type {
	// neither matrix nor scalar - cannot be used
};
template<bool Element_op, class Base>
class tmatrix_type<true, false, Element_op, Base>
	: public tmatrix_dim<Base::rows, Base::columns, Base> {
	typedef tmatrix_dim<Base::rows, Base::columns, Base> m_base;
public:
	// matrix, not scalar
	tmatrix_type() {}
	template<class A> tmatrix_type(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix_type(const A &a, const B &b)
		: m_base(a, b) {}
};
template<bool Element_op, class Base>
class tmatrix_type<true, true, Element_op, Base>
	: public tmatrix_dim<Base::rows, Base::columns, Base> {
	typedef tmatrix_dim<Base::rows, Base::columns, Base> m_base;
public:
	// matrix and sclar
	tmatrix_type() {}
	template<class A> tmatrix_type(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix_type(const A &a, const B &b)
		: m_base(a, b) {}

	operator float() const;
};
template<bool Element_op, class Base>
class tmatrix_type<false, true, Element_op, Base> : public Base {
public:
	// scalar, not matrix
	tmatrix_type() {}
	template<class A> tmatrix_type(const A &a) : Base(a) {}
	template<class A, class B> tmatrix_type(const A &a, const B &b)
		: Base(a, b) {}

	operator float() const;
};

// selects superclasses to provide (only) the supported operations for
// expressions
template<class Base>
class tmatrix : public tmatrix_type<Base::is_matrix, Base::is_scalar,
	Base::element_op, Base> {
	typedef tmatrix_type<Base::is_matrix, Base::is_scalar,
		Base::element_op, Base> m_base;
public:
	tmatrix() {}
	template<class A> tmatrix(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix(const A &a, const B &b)
		: m_base(a, b) {}
};

// friendly matrix
template<int Rows, int Cols = Rows>
class matrix : public tmatrix<basic_matrix<Rows, Cols> > {
public:
	matrix() {}
	matrix(zero_t z) { assign(*this, z); }
	matrix(identity_t i) { assign(*this, i); }
	template<class T> matrix(const tmatrix_dim<Rows, Cols, T> &);
	template<class T> matrix(const tmatrix_ndim<T> &);

	matrix &operator=(zero_t z) { assign(*this, z); return *this; }
	matrix &operator=(identity_t i) { assign(*this, i); return *this; }
	template<class T> matrix &operator=(const tmatrix_dim<Rows, Cols, T> &);
	template<class T> matrix &operator=(const tmatrix_ndim<T> &);
};

// friendly row matrix
template<int Rows>
class vector : public tmatrix<basic_matrix<Rows, 1> > {
public:
	vector() {}
	vector(zero_t z) { assign(*this, z); }
	template<class T> vector(const tmatrix_dim<Rows, 1, T> &);
	template<class T> vector(const tmatrix_ndim<T> &);

	vector &operator=(zero_t z) { assign(*this, z); return *this; }
	template<class T> vector &operator=(const tmatrix_dim<Rows, 1, T> &);
	template<class T> vector &operator=(const tmatrix_ndim<T> &);
};

} // namespace math
} // namespace cpse

#endif
