#include <cpse/math/matrix.tpp>

#include <limits>
#include <cmath>
#include <stdexcept>

namespace cpse {
namespace math {

typedef basic_matrix<4, 4> m4;
typedef basic_matrix<4, 1> v4;

matrix<4> frustum(float l, float r, float b, float t, float n, float f)
{
	matrix<4> dest = zero;

	float n2 = 2.f * n;
	float rl = r - l;
	float tb = t - b;
	float fn = f - n;

	dest(0,0) = n2/rl;
	dest(0,2) = (r+l)/rl;
	dest(1,1) = n2/tb;
	dest(1,2) = (t+b)/tb;
	dest(2,2) = -(f+n)/fn;
	dest(2,3) = -2.f*f*n/fn;
	dest(3,2) = -1.f;

	return dest;
}

matrix<4> ortho(float l, float r, float b, float t, float n, float f)
{
	matrix<4> dest = zero;

	float rl = r - l;
	float tb = t - b;
	float fn = f - n;

	dest(0,0) = 2.f/rl;
	dest(0,3) = (r+l)/rl;
	dest(1,1) = 2.f/tb;
	dest(1,3) = -(t+b)/tb;
	dest(2,2) = -2.f/fn;
	dest(2,3) = (f+n)/fn;
	dest(3,3) = 1.f;

	return dest;
}

void matrix4_assign_handler::go(m4 &d, matrix_binop<binop_mul, m4, m4> s)
{
	for (int i=0; i!=4; ++i) {

		d(i,0) = s.a(i,0)*s.b(0,0) + s.a(i,1)*s.b(1,0)
			+ s.a(i,2)*s.b(2,0) + s.a(i,3)*s.b(3,0);

		d(i,1) = s.a(i,0)*s.b(0,1) + s.a(i,1)*s.b(1,1)
			+ s.a(i,2)*s.b(2,1) + s.a(i,3)*s.b(3,1);

		d(i,2) = s.a(i,0)*s.b(0,2) + s.a(i,1)*s.b(1,2)
			+ s.a(i,2)*s.b(2,2) + s.a(i,3)*s.b(3,2);

		d(i,3) = s.a(i,0)*s.b(0,3) + s.a(i,1)*s.b(1,3)
			+ s.a(i,2)*s.b(2,3) + s.a(i,3)*s.b(3,3);
	}
}

void matrix4_assign_handler::go(v4 &d, matrix_binop<binop_mul, m4, v4> s)
{
	d.x = s.a(0,0)*s.b.x + s.a(0,1)*s.b.y + s.a(0,2)*s.b.z + s.a(0,3)*s.b.w;
	d.y = s.a(1,0)*s.b.x + s.a(1,1)*s.b.y + s.a(1,2)*s.b.z + s.a(1,3)*s.b.w;
	d.z = s.a(2,0)*s.b.x + s.a(2,1)*s.b.y + s.a(2,2)*s.b.z + s.a(2,3)*s.b.w;
	d.w = s.a(3,0)*s.b.x + s.a(3,1)*s.b.y + s.a(3,2)*s.b.z + s.a(3,3)*s.b.w;
}

void assign(m4 &m, identity_t)
{
	m(0,1) = m(0,2) = m(0,3) = 0.f;
	m(1,0) = m(1,2) = m(1,3) = 0.f;
	m(2,0) = m(2,1) = m(2,3) = 0.f;
	m(3,0) = m(3,1) = m(3,2) = 0.f;
	m(0,0) = m(1,1) = m(2,2) = m(3,3) = 1.f;
}

void matrix4_assign_handler::go(m4 &dest, 
	matrix_unop<unop_translation, basic_matrix<3, 1> > m)
{
	dest(0,3) = m.a.x;
	dest(1,3) = m.a.y;
	dest(2,3) = m.a.z;
	dest(0,1) = dest(0,2) = 0.f;
	dest(1,0) = dest(1,2) = 0.f;
	dest(2,0) = dest(2,1) = 0.f;
	dest(3,0) = dest(3,1) = dest(3,2) = 0.f;
	dest(0,0) = dest(1,1) = dest(2,2) = dest(3,3) = 1.f;
}

void matrix4_assign_handler::go(m4 &dest, matrix_unop<unop_scaling, 
	basic_matrix<3, 1> > src)
{
	dest(0,0) = src.a.x;
	dest(1,1) = src.a.y;
	dest(2,2) = src.a.z;
	dest(3,3) = 1.f;
	dest(0,1) = dest(0,2) = dest(0,3) = 0.f;
	dest(1,0) = dest(1,2) = dest(1,3) = 0.f;
	dest(2,0) = dest(2,1) = dest(2,3) = 0.f;
	dest(3,0) = dest(3,1) = dest(3,2) = 0.f;
}

void matrix4_assign_handler::go(m4 &dest, matrix_unop<unop_scaling, 
	float_keeper> src)
{
	dest(0,0) = dest(1,1) = dest(2,2) = src.a.f;
	dest(3,3) = 1.f;
	dest(0,1) = dest(0,2) = dest(0,3) = 0.f;
	dest(1,0) = dest(1,2) = dest(1,3) = 0.f;
	dest(2,0) = dest(2,1) = dest(2,3) = 0.f;
	dest(3,0) = dest(3,1) = dest(3,2) = 0.f;
}

void matrix4_assign_handler::go(m4 &dest, matrix_binop<binop_rotation, 
	basic_matrix<3, 1>, float_keeper> src)
{
	float a = src.b.f;
	float c = cos(a), s = sin(a);
	float u = src.a.x, v = src.a.y, w = src.a.z;

	dest(0,0) =    c + u*u*(1-c);
	dest(0,1) =  w*s + v*u*(1-c);
	dest(0,2) = -v*s + w*u*(1-c);
	dest(1,0) = -w*s + u*v*(1-c);
	dest(1,1) =    c + v*v*(1-c);
	dest(1,2) =  u*s + w*v*(1-c);
	dest(2,0) =  v*s + u*w*(1-c);
	dest(2,1) = -u*s + v*w*(1-c);
	dest(2,2) =    c + w*w*(1-c);

	dest(3,0) = dest(3,1) = dest(3,2) = dest(0,3) = dest(1,3)
		= dest(2,3) = 0.f;
	dest(3,3) = 1.f;
}

void matrix4_assign_handler::go(m4 &dest, matrix_binop<binop_rotation, 
	x_axis_type, float_keeper> src)
{
	float c = cos(src.b.f), s = sin(src.b.f);

	dest(1,1) = +c;
	dest(2,2) = +c;
	dest(1,2) = -s;
	dest(2,1) = +s;

	dest(0,1) = dest(0,2) = 0.f;
	dest(1,0) = dest(2,0) = 0.f;
	dest(0,0) = 1.f;

	dest(0,3) = dest(1,3) = dest(2,3) = dest(3,0) = dest(3,1)
		= dest(3,2) = 0.f;
	dest(3,3) = 1.f;
}

void matrix4_assign_handler::go(m4 &dest, matrix_binop<binop_rotation, 
	y_axis_type, float_keeper> src)
{
	float c = cos(src.b.f), s = sin(src.b.f);

	dest(0,0) = +c;
	dest(2,2) = +c;
	dest(2,0) = -s;
	dest(0,2) = +s;

	dest(0,1) = dest(1,2) = 0.f;
	dest(1,0) = dest(2,1) = 0.f;
	dest(1,1) = 1.f;

	dest(0,3) = dest(1,3) = dest(2,3) = dest(3,0) = dest(3,1)
		= dest(3,2) = 0.f;
	dest(3,3) = 1.f;
}

void matrix4_assign_handler::go(m4 &dest, matrix_binop<binop_rotation, 
	z_axis_type, float_keeper> src)
{
	float c = cos(src.b.f), s = sin(src.b.f);

	dest(0,0) = +c;
	dest(1,1) = +c;
	dest(0,1) = -s;
	dest(1,0) = +s;

	dest(0,2) = dest(1,2) = 0.f;
	dest(2,0) = dest(2,1) = 0.f;
	dest(2,2) = 1.f;

	dest(0,3) = dest(1,3) = dest(2,3) = dest(3,0) = dest(3,1)
		= dest(3,2) = 0.f;
	dest(3,3) = 1.f;
}

float det_bm(const m4 &m)
{
	return m(0,0) * det(cofactor<0,0>(m))
		- m(1,0) * det(cofactor<1,0>(m))
		+ m(2,0) * det(cofactor<2,0>(m))
		- m(3,0) * det(cofactor<3,0>(m));
}

void matrix4_assign_handler::go(m4 &dest, matrix_unop<unop_adjoint, m4> src)
{
	dest(0,0) = +det(cofactor<0,0>(src.a));
	dest(1,0) = -det(cofactor<0,1>(src.a));
	dest(2,0) = +det(cofactor<0,2>(src.a));
	dest(3,0) = -det(cofactor<0,3>(src.a));

	dest(0,1) = -det(cofactor<1,0>(src.a));
	dest(1,1) = +det(cofactor<1,1>(src.a));
	dest(2,1) = -det(cofactor<1,2>(src.a));
	dest(3,1) = +det(cofactor<1,3>(src.a));

	dest(0,2) = +det(cofactor<2,0>(src.a));
	dest(1,2) = -det(cofactor<2,1>(src.a));
	dest(2,2) = +det(cofactor<2,2>(src.a));
	dest(3,2) = -det(cofactor<2,3>(src.a));

	dest(0,3) = -det(cofactor<3,0>(src.a));
	dest(1,3) = +det(cofactor<3,1>(src.a));
	dest(2,3) = -det(cofactor<3,2>(src.a));
	dest(3,3) = +det(cofactor<3,3>(src.a));
}

void matrix4_assign_handler::go(m4 &dest, matrix_unop<unop_inverse, m4> src)
{
	float det_m = det_bm(src.a);
//	if (close(det_m, 0.f))
//		throw std::logic_error("inverse of singular matrix requested");

	float d = 1.f / det_m;

	// dest = adjoint(src.a) / det(src.a)
	dest(0,0) = d * +det(cofactor<0,0>(src.a));
	dest(1,0) = d * -det(cofactor<0,1>(src.a));
	dest(2,0) = d * +det(cofactor<0,2>(src.a));
	dest(3,0) = d * -det(cofactor<0,3>(src.a));

	dest(0,1) = d * -det(cofactor<1,0>(src.a));
	dest(1,1) = d * +det(cofactor<1,1>(src.a));
	dest(2,1) = d * -det(cofactor<1,2>(src.a));
	dest(3,1) = d * +det(cofactor<1,3>(src.a));

	dest(0,2) = d * +det(cofactor<2,0>(src.a));
	dest(1,2) = d * -det(cofactor<2,1>(src.a));
	dest(2,2) = d * +det(cofactor<2,2>(src.a));
	dest(3,2) = d * -det(cofactor<2,3>(src.a));

	dest(0,3) = d * -det(cofactor<3,0>(src.a));
	dest(1,3) = d * +det(cofactor<3,1>(src.a));
	dest(2,3) = d * -det(cofactor<3,2>(src.a));
	dest(3,3) = d * +det(cofactor<3,3>(src.a));
}

} // namespace math
} // namespace cpse
