
namespace cpse {
namespace math {

template<class Op, class A> struct matrix_unop {};

template<class A>
struct matrix_unop<unop_negate, A> {
	static const int rows = A::rows;
	static const int columns = A::columns;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = A::is_scalar;
	static const bool error = A::error || (!is_matrix && !is_scalar);
	static const bool element_op = A::element_op;
	typedef matrix_unop<unop_negate, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
	template<int I> float element() const { return -a.element<I>(); }
};

template<class A>
struct matrix_unop<unop_transpose, A> {
	template<int I> struct index {
		enum { ir = I % A::rows };
		enum { ic = I / A::rows };
		enum { jr = ic };
		enum { jc = ir };
		static const int value = jr + jc*A::rows;
	};
	static const int rows = A::columns;
	static const int columns = A::rows;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = false;
	static const bool error = A::error || !is_matrix;
	static const bool element_op = A::element_op;
	typedef matrix_unop<unop_transpose, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
	template<int I> float element() const
		{ return a.element<index<I>::value>(); }
};

template<class A>
struct matrix_unop<unop_adjoint, A> {
	static const int rows = A::columns;
	static const int columns = A::rows;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = false;
	static const bool error = A::error || !is_matrix;
	static const bool element_op = false;
	typedef matrix_unop<unop_adjoint, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_slice, A> {
	static const int rows = -99;
	static const int columns = -99;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = false;
	static const bool error = A::error || !is_matrix;
	static const bool element_op = false;
	typedef matrix_unop<unop_slice, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_inverse, A> {
	static const int rows = A::rows;
	static const int columns = A::columns;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = false;
	static const bool error = A::error || !is_matrix;
	static const bool element_op = false;
	typedef matrix_unop<unop_inverse, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<int R, int C, class A>
struct matrix_unop<unop_cofactor<R, C>, A> {
	template<int I> struct index {
		enum { ir = I % (A::rows-1) };
		enum { ic = I / (A::rows-1) };
		enum { jr = ir >= R ? ir+1 : ir };
		enum { jc = ic >= C ? ic+1 : ic };
		static const int value = jr + jc*A::rows;
	};
	static const int rows = A::rows-1;
	static const int columns = A::columns-1;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = false;
	static const bool error = A::error || !is_matrix || rows <= 0
		|| columns <= 0;
	static const bool element_op = A::element_op;
	typedef matrix_unop<unop_cofactor<R, C>, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
	template<int I> float element() const
		{ return a.element<index<I>::value>(); }
};

template<class A>
struct matrix_unop<unop_wdiv, A> {
	static const int rows = 3;
	static const int columns = 1;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = false;
	static const bool error = A::error || !is_matrix || A::rows != 4
		|| A::columns != 1;
	static const bool element_op = false;
	typedef matrix_unop<unop_wdiv, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_normalize, A> {
	static const int rows = A::rows;
	static const int columns = 1;
	static const bool is_matrix = A::is_matrix;
	static const bool is_scalar = false;
	static const bool error = A::error || !is_matrix || A::columns != 1;
	static const bool element_op = false;
	typedef matrix_unop<unop_normalize, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_translation, A> {
	static const int rows = 4;
	static const int columns = 4;
	static const bool is_matrix = true;
	static const bool is_scalar = false;
	static const bool error = A::error || !A::is_matrix || A::rows != 3
		|| A::columns != 1;
	static const bool element_op = false;
	typedef matrix_unop<unop_translation, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

template<class A>
struct matrix_unop<unop_scaling, A> {
	static const int rows = 4;
	static const int columns = 4;
	static const bool is_matrix = true;
	static const bool is_scalar = false;
	static const bool error = A::error
		|| (A::is_matrix + A::is_scalar != 1);
	static const bool element_op = false;
	typedef matrix_unop<unop_scaling, A> base_type;
	typename matrix_ref<A>::type a;
	matrix_unop(const A &a_) : a(a_) {}
};

} // namespace math
} // namespace cpse
